/**************************************************
opengate Mac addr auth program

   module for Controlling sessions 

Copyright (C) 2011 Opengate Project Team
Written by Yoshiaki Watanabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Email: watanaby@is.saga-u.ac.jp
**************************************************/
#include "opengatemd.h"

/*******************************
add session for the ip address
add ipfw-rule and table-entry, write db and log 
*******************************/
int addSession(char* ipAddress, char* macAddress, char* userId, char* extraId){
  int ruleNumber;
  static int nSerialError=0; /* number of serial errors for addition */
  int failToAdd=FALSE;

  /* open firewall */
  ruleNumber=OpenClientGate(ipAddress, AVOID_OVERLAP, 
			    userId, extraId, macAddress);

  /* check error */
  if(-2<=ruleNumber && ruleNumber<=2) failToAdd=TRUE;

  /* if previously opened as -ruleNumber */
  else if(ruleNumber<0){
    ruleNumber=-ruleNumber;

    /* if the rule is for opengate, add new rule for opengatemd, else use it */
    if(IsActiveRuleInOpengateWorkDb(ruleNumber)){
      ruleNumber=OpenClientGate(ipAddress, ALLOW_OVERLAP, 
				userId, extraId, macAddress);

      /* check error */
      if(-2<=ruleNumber && ruleNumber<=2) failToAdd=TRUE;
    }
  }

  /* if addtion is failed, retry on next packet (clear cache for checking) */
  /* as to prevent error loop, retry count is limitted */ 
  if( failToAdd || !IsIpAddressFoundInIpfw(ipAddress) ){
    nSerialError++;
    if(nSerialError<1) DelCacheItem("", ipAddress);
    err_msg("ERR at %s#%d: ipfw rule addition is failed(%d-%d)",
	      __FILE__,__LINE__, failToAdd, nSerialError);
    return FALSE;
  }
  nSerialError=0;

  /* write to session db */
  InsertSessionToWorkDb(ipAddress, userId, extraId, macAddress, ruleNumber);

  /* write open log to syslog */
  WriteOpenToSyslog(userId, extraId, ipAddress, macAddress);

  /* write log to management db */
  PutOpenToMngDb(macAddress, ipAddress);
  return TRUE;
}

/*******************************
delete session for the ip address
(del ipfw-rule and table-entry, write db and log) 
*******************************/
void delSession(char* ipAddress){
  char userId[USERMAXLN]="";
  char extraId[USERMAXLN]="";
  char macAddress[ADDRMAXLN]="";
  int openTime=0;
  int ruleNumber=0;
  int checkTime=0;
  int success;

  /* get information from session table in work db */
  success=GetSessionFromWorkDb(ipAddress, userId, extraId, &openTime, 
				 &checkTime, macAddress, &ruleNumber);

  /* close firewall and refresh the cache */
  if(success) CloseClientGate(ruleNumber);

  /* del from cache */
  DelCacheItem("", ipAddress);
  
  /* del from session from session table in work db */
  DelSessionFromWorkDb(ipAddress);

  /* write close log to syslog */
  WriteCloseToSyslog(userId, extraId, ipAddress, macAddress, openTime);
  if(debug>0) WriteSessionInfoToSyslog(userId, extraId, ipAddress, 
				       macAddress, ruleNumber);

  /* write close log to management db */
  PutCloseToMngDb(ipAddress);
}

/*******************************
renew session for the ip address
(renew time in table entry) 
*******************************/
void renewSession(char* ipAddress){

  /* renew the check time */
  UpdateCheckTimeInWorkDb(ipAddress);
}

/*******************************
search db to find ip that is not used for a long time
del ipfw rule, del table entry, write db/log 
*******************************/
void delUselessSessions(void){

  /* scan db to remove old sessions */
  DelUselessSessionsInWorkDb(TRUE);

  /* scan ipfw rule to close unmatched db row */
  CloseUnmatchSessions();
}

/**********************************
CALLBACK FUNCTION
called back from sqlite3_exec 
in sessiondb.c/delUselessSessionsInDb & delAllSessionsInDb
**********************************/
int closeSession(void* pParam, int argc, char *argv[], char* colName[]){

  int ruleNumber;
  char* userId;
  char* extraId;
  char* ipAddress;
  char* macAddress;
  int openTime;

  if(argc < 5) return 1; /* SQLITE_ERROR */
  
  ruleNumber=atoi(argv[0]);
  userId=argv[1];
  extraId=argv[2];
  ipAddress=argv[3];
  macAddress=argv[4];
  openTime=atoi(argv[5]);

  /* close firewall */
  CloseClientGate(ruleNumber);
  DelCacheItem(macAddress, ipAddress);

  /* write close log to syslog */
  WriteCloseToSyslog(userId, extraId, ipAddress, macAddress, openTime);
  if(debug>0) WriteSessionInfoToSyslog(userId, extraId, ipAddress, 
				       macAddress, ruleNumber);

  /* write close log to management db */
  PutCloseToMngDb(ipAddress);
  return 0; /* SQLITE_OK */
}

/*******************************
search db to find all ip
del ipfw rule, del table entry, write db/log 
*******************************/
void delAllSessions(void){

  /* scan db to remove all sessions. argument indicates delayed flag */
  DelUselessSessionsInWorkDb(FALSE);

  /* scan ipfw rule to close unmatched db row */
  CloseUnmatchSessions();
}

/**********************************************
is the same mac/ip session found 
**********************************************/
int isMatchedSessionFound(char* ipAddress, char* macAddressNow){
  int ret;
  char userId[USERMAXLN];
  char extraId[USERMAXLN];
  int openTime;
  int checkTime;
  char macAddress[ADDRMAXLN];
  int ruleNumber;

  /* get info for the ipAddress */
  ret=GetSessionFromWorkDb(ipAddress, userId, extraId, 
			 &openTime, &checkTime, macAddress,
			     &ruleNumber);
  if(ret==FALSE) return FALSE;

  /* if detected mac and saved mac is matched, return true */
  if(strncmp(macAddress, macAddressNow, ADDRMAXLN)==0)return TRUE;

  /* if exist but unmatched pair, delete the session and return false */
  else{
    DelSession(ipAddress);
    return FALSE;
  }
}

/************************************
debug routine for hash table
************************************/
void dumpTable(DB* table){
  DBT hashKey;
  DBT hashVal;
  int ruleNumber=0;
  char ipAddress[ADDRMAXLN]="?";
  int ret;

  memset(&hashKey, 0, sizeof(DBT));
  memset(&hashVal, 0, sizeof(DBT));
  ret=table->seq(table, &hashKey, &hashVal, R_FIRST);
  while(ret==0){

    ruleNumber=*(int*)(hashVal.data);
    strncpy(ipAddress, (char*)hashKey.data, ADDRMAXLN);
    err_msg("%d:%s", ruleNumber, ipAddress);

    /* get next entry */
    ret=table->seq(table, &hashKey, &hashVal, R_NEXT);
  }
}

/************************************************
 close sessions that lost ipfw rule or database entry
( load rules from ipfw and database, then compare ) 
  sessionTable    sessionTable      ipfw
 for opengatemd   for opengate   ruleTable
   key=IpAddr       IpAddr        IpAddr
   val=0           ProcessID     IpfwRule
      ^              ^               ^
      |--------------|               |
             |----- compare key -----|
entry in ruleTable must be found in one of the sessionTables
if exists in ruleTable and not in sessionTable, remove rule
if exists in sessionTable and not in ruleTable, remove session
caution: opengate saves only one of ipv6 addresses in db.
       : the ruleNumber for opengate is stored with negative value
*************************************************/
void closeUnmatchSessions(void){
  DB* ruleTable;
  DB* sessionTable;

  /* prepare hash table for active sessions */
  if((sessionTable = dbopen(NULL, O_CREAT | O_RDWR, 0644, DB_HASH, NULL)) == NULL) {
    err_msg("ERR at %s#%d: fail to open hash table",__FILE__,__LINE__);
    terminateProg(0);
  }

  /* prepare hash table for rule numbers */
  if((ruleTable = dbopen(NULL, O_CREAT | O_RDWR, 0644, DB_HASH, NULL)) == NULL) {
    err_msg("ERR at %s#%d: fail to open hash table",__FILE__,__LINE__);
    terminateProg(0);
  }

  /* get rule numbers from ipfw */
  GetRuleTableFromIpfw(ruleTable);

  /* get active sessions from work db for opengateMd and opengate */
  GetSessionTableFromWorkDb(sessionTable);
  GetSessionTableFromOpengateWorkDb(sessionTable);

  /****debug print
  err_msg("Rule");
  dumpTable(ruleTable);
  err_msg("Session");
  dumpTable(sessionTable);
  ****/

  /* remove unmatched entry */
  RemoveSessionUnmatchedToIpfwRule(ruleTable, sessionTable);
  RemoveIpfwRuleUnmatchedToSession(ruleTable, sessionTable);

  /* erase table area */
  sessionTable->close(sessionTable);
  ruleTable->close(ruleTable);
}

/*******************************
remove session in db unmatched to ipfw active rule
 (db entry exists but no ipfw rule is found)
*******************************/
void removeSessionUnmatchedToIpfwRule(DB* ruleTable, DB* sessionTable){
  DBT hashKey;
  DBT hashVal;
  int retRuleTbl;
  int retSesTbl;
  int pid;
  int ruleNumber;
  char ipAddress[ADDRMAXLN];
  char userId[USERMAXLN];
  char extraId[USERMAXLN];
  char macAddress[ADDRMAXLN];
  int openTime;
  char openTimeStr[WORDMAXLN];
  int checkTime;

  /* scan session table to find entry unmatched to rule table */
  /* get first entry of session table */
  memset(&hashKey, 0, sizeof(DBT));
  memset(&hashVal, 0, sizeof(DBT));
  retSesTbl=sessionTable->seq(sessionTable, &hashKey, &hashVal, R_FIRST);
  while(retSesTbl==0){

    /* (get-function returns 0 on success) */
    /* if found a session, save the data */
    /* if pid>0, the session is controlled by opengate process having pid */
    /* if pid=0, the session is controlled by opengatemd */
    pid = *(int*)(hashVal.data);
    strncpy(ipAddress, (char*)hashKey.data, ADDRMAXLN);

    /* get entry having same key in rule table */
    hashKey.data = ipAddress;
    hashKey.size = strlen(ipAddress)+1;
    memset(&hashVal, 0, sizeof(DBT));
    retRuleTbl=ruleTable->get(ruleTable, &hashKey, &hashVal, 0);

    /*** rule does not exist and the session is controlled by opengatemd */
    /* (get-function returns 0 on success) */
    if(retRuleTbl!=0 && pid==0){
      
      /* write log and close session */
      if(!GetSessionFromWorkDb(ipAddress, userId, extraId, 
			      &openTime, &checkTime, macAddress, &ruleNumber)){
	err_msg("ERR at %s#%d: fail to get session info",__FILE__,__LINE__);
      }else{
	WriteCloseToSyslog(userId, extraId, ipAddress, macAddress, openTime);
	PutCloseToMngDb(ipAddress);
	DelSessionFromWorkDb(ipAddress);
      }

      /* write session info to syslog */
      if(debug>0) WriteSessionInfoToSyslog(userId, extraId, ipAddress, 
					   macAddress, ruleNumber);
    }

    /*** rule does not exist and the session is controlled by opengate */
    else if(retRuleTbl!=0 && pid>0){
	
      /* write log and close session */
      if(!GetSessionFromOpengateWorkDb(ipAddress, userId, 
				      openTimeStr, macAddress, &ruleNumber)){
	err_msg("ERR at %s#%d: fail to get session info",__FILE__,__LINE__);
      }else{
	WriteCloseToSyslog(userId, "", ipAddress, macAddress, time(NULL));
	SetCloseTimeToOpengateWorkDb(ipAddress);
      }

      /* write session info to syslog */
      if(debug>0) WriteSessionInfoToSyslog(userId, "", ipAddress, 
					   macAddress, ruleNumber);
    }
    
    /*** rule exists and the session is controlled by opengate */
    else if(retRuleTbl==0 && pid>0){

      /* if the controlling opengate process (pid) does not exist */
      /* (kill(pid,0)==0||EPERM==errno) means existence of pid process */
      if( !(kill(pid,0)==0||EPERM==errno) ){

	/* if other process uses same rule, then hold the rule, else del. */
	ruleNumber = *(int*)(hashVal.data);
	if(!IsProcessFoundForTheRule(ruleNumber)){
	  CloseClientGate(ruleNumber);
	}

	/* remove cache and session */
	DelCacheItem("", ipAddress);
	if(!GetSessionFromOpengateWorkDb(ipAddress, userId, 
					openTimeStr, macAddress, &ruleNumber)){
	  err_msg("ERR at %s#%d: fail to get session info",__FILE__,__LINE__);
	}else{
	  WriteCloseToSyslog(userId, "", ipAddress, macAddress, time(NULL));
	  SetCloseTimeToOpengateWorkDb(ipAddress);
	}

	/* write session info to syslog */
	if(debug>0) WriteSessionInfoToSyslog(userId, extraId, ipAddress, 
					     macAddress, ruleNumber);
      }
    }

    /*** rule exists and the session is controlled by opengatemd */
    else{
      ; /* do nothing */
    }

    /* get next rule entry */
    retSesTbl=sessionTable->seq(sessionTable, &hashKey, &hashVal, R_NEXT);
  }
}

/***************************************************
find process using the ipfw rule
 process for opengate: search by 'ps ax' command
 process for opengatemd: search by sqlite db
***************************************************/
int isProcessFoundForTheRule(int ruleNumber){
  FILE *fpipe;
  char cgiName[WORDMAXLN];
  char ruleNumberStr[WORDMAXLN];
  char buf[BUFFMAXLN];
  int found=FALSE;
  char* p;

  /*** search the process for opengate by using 'ps ax' */
  /* 'ps ax' returns output including rule numbers as */
  /* 12345 ?? S 0:00.05 opengatesrv.cgi: user001, */
  /* [192.168.1.100(10008)],[1234:5678::100(10010)], */

  /* prepare search string */
  strncpy(cgiName, GetConfValue("MainCgi"), WORDMAXLN);
  snprintf(ruleNumberStr, WORDMAXLN, "(%d)]",ruleNumber);

  /* exec 'ps ax' and open pipe */
  if((fpipe=Popenl(0, "r", "ps","ax",(char *)0)) == NULL){ 
      err_msg("ERR at %s#%d: exec ps ax error",__FILE__,__LINE__);
  }

  /* find above strings */
  while(fgets(buf, BUFFMAXLN, fpipe)!=NULL){
    if((p=strstr(buf, cgiName))==NULL) continue; /* not cgi */
    if(strstr(p, ruleNumberStr)==NULL) continue; /* not rule */
    found=TRUE;
    break;
  }
  
  /* close pipe */
  Pclose(fpipe);
  
  /* if found, return true */
  if(found) return TRUE;

  /*** search the rule number for opengatemd by database */
  return IsActiveRuleInWorkDb(ruleNumber);
}

/***********************************
remove active ipfw rule unmatched to session in db
 (ipfw rule exists but no db entry is found)
***********************************/
void  removeIpfwRuleUnmatchedToSession(DB* ruleTable, DB* sessionTable){
  DBT hashKey;
  DBT hashVal;
  int retRuleTbl;
  int retSesTbl;
  int ruleNumber;
  char ipAddress[ADDRMAXLN];

  /* scan ipfw rule table to find entry unmatched to session table */
  /* get first entry of ipfw rule table */
  memset(&hashKey, 0, sizeof(DBT));
  memset(&hashVal, 0, sizeof(DBT));
  retRuleTbl=ruleTable->seq(ruleTable, &hashKey, &hashVal, R_FIRST);
  while(retRuleTbl==0){

    /* if found in rule table, save data */
    ruleNumber=*(int*)(hashVal.data);
    strncpy(ipAddress, (char*)hashKey.data, ADDRMAXLN);

    /* get entry having same key in session table */
    hashKey.data = ipAddress;
    hashKey.size = strlen(ipAddress)+1;
    memset(&hashVal, 0, sizeof(DBT));
    retSesTbl=sessionTable->get(sessionTable, &hashKey, &hashVal, 0);

    /* if not found in session table, remove entry in ipfw rule */
    if(retSesTbl!=0){
      
      /* opengate cannot save all ipv6 address. */
      /* if the address is ipv6 */
      /* and if the rule number is found in openate db, */
      /* don't remove it */
      if((strchr(ipAddress, ':')!=NULL) &&
	 IsActiveRuleInOpengateWorkDb(ruleNumber)){
	; /* do nothing */
      }else{
	
	/* remove ipfw rule */
	CloseClientGate(ruleNumber);
	DelCacheItem("", ipAddress);

	/* write log */
	WriteCloseToSyslog("?", "", ipAddress, "?", time(NULL));
	PutCloseToMngDb(ipAddress);

	/* write session info to syslog */
	if(debug>0) WriteSessionInfoToSyslog("?", "", ipAddress, 
					     "?", ruleNumber);
      }
    }
    /* get next rule entry */
    retRuleTbl=ruleTable->seq(ruleTable, &hashKey, &hashVal, R_NEXT);
  }
}


/******************************************
write open message to syslog
******************************************/
void writeOpenToSyslog(char* userId, char* extraId, char* ipAddress, char* macAddress){

  if(extraId[0]=='\0'){
    err_msg("OPEN: user %s from %s at %s", 
	    userId, ipAddress, macAddress);
  }else{
    err_msg("OPEN: user %s%s%s from %s at %s", 
	    userId, GetConfValue("UserIdSeparator"), extraId, 
	    ipAddress, macAddress);
  }
}

/******************************************
write close message to syslog
******************************************/
void writeCloseToSyslog(char* userId, char* extraId, char* ipAddress, char* macAddress, int openTime){

  double time_l;
  int hour, min, sec;

  time_l=difftime((int)time(NULL), openTime);
  hour=time_l/60/60;
  min=(time_l-hour*60*60)/60;
  sec=(time_l-hour*60*60-min*60);

  if(extraId[0]=='\0'){
    err_msg("CLOS: user %s from %s at %s ( %02d:%02d:%02d )", 
	    userId, ipAddress, macAddress, hour,min,sec);
  }else{
    err_msg("CLOS: user %s%s%s from %s at %s ( %02d:%02d:%02d )", 
	    userId, GetConfValue("UserIdSeparator"), extraId, 
	    ipAddress, macAddress, hour,min,sec);
  }
}

/******************************************
write session info message to syslog
******************************************/
void writeSessionInfoToSyslog(char* userId, char* extraId, char* ipAddress, char* macAddress, int ruleNumber){

  char detectTimeStr[WORDMAXLN];
  int ttl;

  /* get client info from macinfo table in workdb */
  GetMacInfoFromWorkDb(macAddress, detectTimeStr, &ttl);

  if(extraId[0]=='\0'){
    err_msg("INFO: user=%s ipaddr=%s macaddr=%s "
	    "ipfwrule=%d ttl=%d lastcheck=%s", 
	    userId, ipAddress, macAddress, ruleNumber, ttl, detectTimeStr);
  }else{
    err_msg("INFO: user=%s%s%s ipaddr=%s macaddr=%s "
	    "ipfwrule=%d ttl=%d lastcheck=%s", 
	    userId, GetConfValue("UserIdSeparator"), extraId, 
	    ipAddress, macAddress, ruleNumber, ttl, detectTimeStr);
  }
}


/**************************************************
 routines for debugging output
 *************************************************/

int AddSession(char* ipAddress, char* macAddress, char* userId, char* extraId){
  int ret;
  if(debug>1) err_msg("DEBUG:=>addSession(%s,%s,%s,%s)",
		      ipAddress, macAddress, userId, extraId);
  ret = addSession(ipAddress, macAddress, userId, extraId);
  if(debug>1) err_msg("DEBUG:(%d)<=addSession( )",ret);
  return ret;
}
void DelSession(char* ipAddress){
  if(debug>1) err_msg("DEBUG:=>delSession(%s)",ipAddress);
  delSession(ipAddress);
  if(debug>1) err_msg("DEBUG:<=delSession( )");
}
void RenewSession(char* ipAddress){
  if(debug>1) err_msg("DEBUG:=>renewSession(%s)",ipAddress);
  renewSession(ipAddress);
  if(debug>1) err_msg("DEBUG:<=renewSession( )");
}
void DelUselessSessions(void){
  if(debug>1) err_msg("DEBUG:=>delUselessSessions()");
  delUselessSessions();
  if(debug>1) err_msg("DEBUG:<=delUselessSessions( )");
}
void DelAllSessions(void){
  if(debug>1) err_msg("DEBUG:=>delAllSessions()");
  delAllSessions();
  if(debug>1) err_msg("DEBUG:<=delAllSessions( )");
}
int CloseSession(void* pParam, int argc, char *argv[], char* colName[]){
  int ret;
  if(debug>1) err_msg("DEBUG:=>closeSession()");
  ret = closeSession(pParam, argc, argv, colName);
  if(debug>1) err_msg("DEBUG:(%d)<=closeSession( )",ret);
  return ret;
}
int IsMatchedSessionFound(char* ipAddress, char* macAddress){
  int ret;
  if(debug>1) err_msg("DEBUG:=>isMatchedSessionFound(%s,%s)",
		      ipAddress, macAddress);
  ret = isMatchedSessionFound(ipAddress, macAddress);
  if(debug>1) err_msg("DEBUG:(%d)<=isMatchedSessionFound( )",ret);
  return ret;
}

void CloseUnmatchSessions(void){
  if(debug>1) err_msg("DEBUG:=>closeUnmatchSessions( )");
  closeUnmatchSessions();
  if(debug>1) err_msg("DEBUG:<=closeUnmatchSessions( )");
}

void WriteOpenToSyslog(char* userId, char* extraId, char* ipAddress, char* macAddress){
  if(debug>1) err_msg("DEBUG:=>writeOpenToSyslog(%s,%s,%s,%s)", userId, extraId, ipAddress, macAddress);
  writeOpenToSyslog(userId, extraId, ipAddress, macAddress);
  if(debug>1) err_msg("DEBUG:<=writeOpenToSyslog( )");
}

void WriteCloseToSyslog(char* userId, char* extraId, char* ipAddress, char* macAddress, int openTime){
  if(debug>1) err_msg("DEBUG:=>writeCloseToSyslog(%s,%s,%s,%s,%d)", userId, extraId, ipAddress, macAddress, openTime);
  writeCloseToSyslog(userId, extraId, ipAddress, macAddress, openTime);
  if(debug>1) err_msg("DEBUG:<=writeCloseToSyslog( )");
}

void WriteSessionInfoToSyslog(char* userId, char* extraId, char* ipAddress, char* macAddress, int ruleNumber){
  if(debug>1) err_msg("DEBUG:=>writeSessionInfoToSyslog(%s,%s,%s,%s,%d)", userId, extraId, ipAddress, macAddress, ruleNumber);
  writeSessionInfoToSyslog(userId, extraId, ipAddress, macAddress, ruleNumber);
  if(debug>1) err_msg("DEBUG:<=writeSessionInfoToSyslog( )");
}


void RemoveSessionUnmatchedToIpfwRule(DB* ruleTable, DB* sessionTable){
  if(debug>1) err_msg("DEBUG:=>removeSessionUnmatchedToIpfwRule()");
  removeSessionUnmatchedToIpfwRule(ruleTable, sessionTable);
  if(debug>1) err_msg("DEBUG:<=removeSessionUnmatchedToIpfwRule()");
}

void RemoveIpfwRuleUnmatchedToSession(DB* ruleTable, DB* sessionTable){
  if(debug>1) err_msg("DEBUG:=>removeIpfwRuleUnmatchedToSession()");
  removeIpfwRuleUnmatchedToSession(ruleTable, sessionTable);
  if(debug>1) err_msg("DEBUG:<=removeIpfwRuleUnmatchedToSession()");

}

int IsProcessFoundForTheRule(int ruleNumber){
  int ret;
  if(debug>1) err_msg("DEBUG:=>isProcessFoundForTheRule(%d)", ruleNumber);
  ret = isProcessFoundForTheRule(ruleNumber);
  if(debug>1) err_msg("DEBUG:<=isProcessFoundForTheRule()");
  return ret;
}

/********************testmain********************
void testmain(){

  char* ipAddress="192.168.0.120";
  char* macAddress="00:01:02:03:04:05";
  char* userId="watanaby";
  char* extraId="";

  addSession(ipAddress, macAddress, userId, extraId);
  sleep(10);
  renewSession(ipAddress);
  sleep(10);
  delUselessSessions();
  sleep(10);
  delUselessSessions();
  delSession(ipAddress);
}
************************************************/
